/*
 * ValueOperationAbstraction.hpp
 *
 *  Created on: 11. 7. 2017
 *	  Author: Jan Travnicek
 */

#ifndef _VALUE_OPERATION_ABSTRACTION_HPP_
#define _VALUE_OPERATION_ABSTRACTION_HPP_

#include <alib/tuple>
#include <alib/memory>
#include <alib/typeindex>

#include <abstraction/OperationAbstraction.hpp>
#include <abstraction/ValueProvider.hpp>

#include <common/ParamQualifiers.hpp>
#include <common/AbstractionHelpers.hpp>

namespace abstraction {

class UnspecifiedType {
};

template < class ReturnType >
class ValueOperationAbstraction : public OperationAbstraction, public ValueProvider < ReturnType >, public ValueProvider < ReturnType & >, public ValueProvider < const ReturnType & >, public ValueProvider < ReturnType && >, public ValueProvider < const ReturnType && > {
protected:
	virtual bool isConst ( ) const override {
		return false;
	}

	virtual bool isRvalueRef ( ) const override {
		return false;
	}

	virtual bool isLvalueRef ( ) const override {
		return false;
	}

	virtual ReturnType & getData ( ) const override {
		return m_data.value ( );
	}

	virtual const ReturnType & getConstData ( ) const override {
		return m_data.value ( );
	}

	mutable std::optional < ReturnType > m_data;

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! cached ( ) )
			m_data = abstraction::apply < ParamTypes ... > ( callback, inputs );
	}

	virtual ext::type_index getReturnTypeIndex ( ) const override {
		return ext::type_index ( typeid ( ReturnType ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < ReturnType > ( );
	}

	virtual bool cached ( ) const override {
		return ( bool ) m_data;
	}

	virtual void reset ( ) override {
		m_data.reset ( );
	}
};

template < class ReturnType >
class ValueOperationAbstraction < const ReturnType > : public OperationAbstraction, public ValueProvider < ReturnType >, public ValueProvider < const ReturnType & >, public ValueProvider < const ReturnType && > {
protected:
	virtual bool isConst ( ) const override {
		return true;
	}

	virtual bool isRvalueRef ( ) const override {
		return false;
	}

	virtual bool isLvalueRef ( ) const override {
		return false;
	}

	virtual ReturnType & getData ( ) const override {
		return m_data.value ( );
	}

	virtual const ReturnType & getConstData ( ) const override {
		return m_data.value ( );
	}

	mutable std::optional < ReturnType > m_data;

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! cached ( ) )
			m_data = abstraction::apply < ParamTypes ... > ( callback, inputs );
	}

	virtual ext::type_index getReturnTypeIndex ( ) const override {
		return ext::type_index ( typeid ( ReturnType ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < const ReturnType > ( );
	}

	virtual bool cached ( ) const override {
		return ( bool ) m_data;
	}

	virtual void reset ( ) override {
		m_data.reset ( );
	}
};

template < class ReturnType >
class ValueOperationAbstraction < ReturnType & > : public OperationAbstraction, public ValueProvider < ReturnType >, public ValueProvider < ReturnType & >, public ValueProvider < const ReturnType & >, public ValueProvider < ReturnType && >, public ValueProvider < const ReturnType && > {
protected:
	virtual bool isConst ( ) const override {
		return false;
	}

	virtual bool isRvalueRef ( ) const override {
		return true;
	}

	virtual bool isLvalueRef ( ) const override {
		return false;
	}

	virtual ReturnType & getData ( ) const override {
		return m_data->get ( );
	}

	virtual const ReturnType & getConstData ( ) const override {
		return m_data->get ( );
	}

	mutable std::optional < std::reference_wrapper < ReturnType > > m_data;

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! cached ( ) )
			m_data = std::reference_wrapper < ReturnType > ( abstraction::apply < ParamTypes ... > ( callback, inputs ) );
	}

	virtual ext::type_index getReturnTypeIndex ( ) const override {
		return ext::type_index ( typeid ( ReturnType ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < ReturnType & > ( );
	}

	virtual bool cached ( ) const override {
		return ( bool ) m_data;
	}

	virtual void reset ( ) override {
		m_data.reset ( );
	}
};

template < class ReturnType >
class ValueOperationAbstraction < const ReturnType & > : public OperationAbstraction, public ValueProvider < ReturnType >, public ValueProvider < const ReturnType & >, public ValueProvider < const ReturnType && > {
protected:
	virtual bool isConst ( ) const override {
		return true;
	}

	virtual bool isRvalueRef ( ) const override {
		return true;
	}

	virtual bool isLvalueRef ( ) const override {
		return false;
	}

	virtual ReturnType & getData ( ) const override {
		return const_cast < ReturnType & > ( m_data->get ( ) );
	}

	virtual const ReturnType & getConstData ( ) const override {
		return m_data->get ( );
	}

	mutable std::optional < std::reference_wrapper < const ReturnType > > m_data;

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! cached ( ) )
			m_data = std::reference_wrapper < const ReturnType > ( abstraction::apply < ParamTypes ... > ( callback, inputs ) );
	}

	virtual ext::type_index getReturnTypeIndex ( ) const override {
		return ext::type_index ( typeid ( ReturnType ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < const ReturnType & > ( );
	}

	virtual bool cached ( ) const override {
		return ( bool ) m_data;
	}

	virtual void reset ( ) override {
		m_data.reset ( );
	}
};

template < class ReturnType >
class ValueOperationAbstraction < ReturnType && > : public OperationAbstraction, public ValueProvider < ReturnType >, public ValueProvider < const ReturnType & >, public ValueProvider < ReturnType && >, public ValueProvider < const ReturnType && > {
protected:
	virtual bool isConst ( ) const override {
		return false;
	}

	virtual bool isRvalueRef ( ) const override {
		return false;
	}

	virtual bool isLvalueRef ( ) const override {
		return true;
	}

	virtual ReturnType & getData ( ) const override {
		return m_data->get ( );
	}

	virtual const ReturnType & getConstData ( ) const override {
		return m_data->get ( );
	}

	mutable std::optional < std::reference_wrapper < ReturnType > > m_data;

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! cached ( ) ) {
			ReturnType && res = abstraction::apply < ParamTypes ... > ( callback, inputs );
			m_data = std::reference_wrapper < ReturnType > ( res );
		}
	}

	virtual ext::type_index getReturnTypeIndex ( ) const override {
		return ext::type_index ( typeid ( ReturnType ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < ReturnType && > ( );
	}

	virtual bool cached ( ) const override {
		return ( bool ) m_data;
	}

	virtual void reset ( ) override {
		m_data.reset ( );
	}
};

template < class ReturnType >
class ValueOperationAbstraction < const ReturnType && > : public OperationAbstraction, public ValueProvider < ReturnType >, public ValueProvider < const ReturnType & >, public ValueProvider < const ReturnType && > {
protected:
	virtual bool isConst ( ) const override {
		return true;
	}

	virtual bool isRvalueRef ( ) const override {
		return false;
	}

	virtual bool isLvalueRef ( ) const override {
		return true;
	}

	virtual ReturnType & getData ( ) const override {
		return const_cast < ReturnType & > ( m_data->get ( ) );
	}

	virtual const ReturnType & getConstData ( ) const override {
		return m_data->get ( );
	}

	mutable std::optional < std::reference_wrapper < const ReturnType > > m_data;

public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		if ( ! cached ( ) ) {
			const ReturnType && res = abstraction::apply < ParamTypes ... > ( callback, inputs );
			m_data = std::reference_wrapper < const ReturnType > ( res );
		}
	}

	virtual ext::type_index getReturnTypeIndex ( ) const override {
		return ext::type_index ( typeid ( ReturnType ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < const ReturnType && > ( );
	}

	virtual bool cached ( ) const override {
		return ( bool ) m_data;
	}

	virtual void reset ( ) override {
		m_data.reset ( );
	}
};

template < >
class ValueOperationAbstraction < void > : public OperationAbstraction {
public:
	template < typename ... ParamTypes, typename Callable >
	inline void run_helper ( Callable callback, const ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) > & inputs ) {
		abstraction::apply < ParamTypes ... > ( callback, inputs );
	}

	virtual ext::type_index getReturnTypeIndex ( ) const override {
		return ext::type_index ( typeid ( void ) );
	}

	ext::set < abstraction::ParamQualifiers::ParamQualifier > getReturnTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < void > ( );
	}

	virtual bool cached ( ) const override {
		return false;
	}

	virtual void reset ( ) override {
	}
};

} /* namespace abstraction */

#endif /* _VALUE_OPERATION_ABSTRACTION_HPP_ */
